/***************************************************************************

    file                 : bt.cpp
    created              : Wed Jan 8 18:31:16 CET 2003
    copyright            : (C) 2002-2004 Bernhard Wymann
    email                : berniw@bluewin.ch
    version              : $Id: bt.cpp,v 1.5.2.2 2008/11/09 17:50:19 berniw Exp $

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef _WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <tgf.h>
#include <track.h>
#include <car.h>
#include <raceman.h>
#include <robottools.h>
#include <robot.h>

#include "driver.h"

#define NBBOTS 10

static const char* botname[NBBOTS] = {
	"bt 1", "bt 2", "bt 3", "bt 4", "bt 5",
	"bt 6", "bt 7", "bt 8", "bt 9", "bt 10"
};

static const char* botdesc[NBBOTS] = {
	"bt 1", "bt 2", "bt 3", "bt 4", "bt 5",
	"bt 6", "bt 7", "bt 8", "bt 9", "bt 10"
};

static Driver *driver[NBBOTS];

static void initTrack(int index, tTrack* track, void *carHandle, void **carParmHandle, tSituation *s);
static void newRace(int index, tCarElt* car, tSituation *s);
static void drive(int index, tCarElt* car, tSituation *s);
static int pitcmd(int index, tCarElt* car, tSituation *s);
static void shutdown(int index);
static int InitFuncPt(int index, void *pt);
static void endRace(int index, tCarElt *car, tSituation *s);


// Module entry point.
extern "C" int bt(tModInfo *modInfo)
{
	int i;
	
	// Clear all structures.
	memset(modInfo, 0, 10*sizeof(tModInfo));

	for (i = 0; i < NBBOTS; i++) {
		modInfo[i].name    = strdup(botname[i]);	// name of the module (short).
		modInfo[i].desc    = strdup(botdesc[i]);	// Description of the module (can be long).
		modInfo[i].fctInit = InitFuncPt;			// Init function.
		modInfo[i].gfId    = ROB_IDENT;				// Supported framework version.
		modInfo[i].index   = i;						// Indices from 0 to 9.
	}
	return 0;
}


// Module interface initialization.
static int InitFuncPt(int index, void *pt)
{
	tRobotItf *itf = (tRobotItf *)pt;

	// Create robot instance for index.
	driver[index] = new Driver(index);
	itf->rbNewTrack = initTrack;	// Give the robot the track view called.
	itf->rbNewRace  = newRace;		// Start a new race.
	itf->rbDrive    = drive;		// Drive during race.
	itf->rbPitCmd   = pitcmd;		// Pit commands.
	itf->rbEndRace  = endRace;		// End of the current race.
	itf->rbShutdown = shutdown;		// Called before the module is unloaded.
	itf->index      = index;		// Index used if multiple interfaces.
	return 0;
}


// Called for every track change or new race.
static void initTrack(int index, tTrack* track, void *carHandle, void **carParmHandle, tSituation *s)
{
	std::ofstream myfile;
	myfile.open ("tracksegments.txt");

	tTrackSeg *curseg = track->seg;

	for(int i=0; i<track->nseg; i++)
	{
		myfile << curseg->vertex[0].x <<" "<< curseg->vertex[0].y <<" "<< curseg->vertex[0].z <<" "<< curseg->vertex[1].x 
			<<" "<< curseg->vertex[1].y <<" "<< curseg->vertex[1].z <<" " <<curseg->radius <<" " << curseg->type << "\n";
		curseg = curseg->next;
	}

	myfile.close();
	path_finder();

	driver[index]->initTrack(track, carHandle, carParmHandle, s);
}


// Start a new race.
static void newRace(int index, tCarElt* car, tSituation *s)
{
	driver[index]->newRace(car, s);
}


// Drive during race.
static void drive(int index, tCarElt* car, tSituation *s)
{
	driver[index]->drive(s);
}


// Pitstop callback.
static int pitcmd(int index, tCarElt* car, tSituation *s)
{
	return driver[index]->pitCommand(s);
}


// End of the current race.
static void endRace(int index, tCarElt *car, tSituation *s)
{
	driver[index]->endRace(s);
}


// Called before the module is unloaded.
static void shutdown(int index)
{
	delete driver[index];
}

void path_finder()
{
	float s = 0.0;
	float p = 0.0;
	float x = 0.0;
	float path_point_x[370];
	float path_point_y[370];
	float X_left = 0.0;
	float Y_left = 0.0;
	float X_right = 0.0;
	float Y_right = 0.0;
	float m1 = 0.0;
	float m2 = 0.0;
	float c1 = 0.0;
	float c2 = 0.0;
	float intersection_X = 0.0;
	float intersection_Y = 0.0;
	float temp1 = 0.0;
	float temp2 = 0.0;
	float temp3 = 0.0;
	float temp4 = 0.0;
	float temp5 = 0.0;
	float distance = 0.0;
	float distance_l = 0.0;
	float distance_r = 0.0;
	float vector_xl = 0.0;
	float vector_yl = 0.0;
	float vector_xr = 0.0;
	float vector_yr = 0.0;
	int i = 0;
	int j = 0;
	int k = 0;

	float x1 = 0.0;
	float x2 = 0.0;
	float y1 = 0.0;
	float y2 = 0.0;
	float dx1 = 0.0;
	float dy1 = 0.0;
	float x3 = 0.0;
	float x4 = 0.0;
	float y3 = 0.0;
	float y4 = 0.0;
	float dx2 = 0.0;
	float dy2 = 0.0;
	float width = 0.0;
	int clip_left = 0;
	int clip_right = 0;

	float Energy = 0.0;
	float minEnergy = 0.0;
	
	SegDetails segment[370];
	
	std::ifstream infile;
	infile.open ("tracksegments.txt");
	for(i = 0; i < 371; i++)
	{
		infile >> segment[i].left.x >> segment[i].left.y >> segment[i].left.z >> segment[i].right.x >> segment[i].right.y >> segment[i].right.z >> segment[i].radius >> segment[i].type;
	}
	infile.close();

	

	/*
	std::ofstream check;
	check.open ("tracksegments_check.txt");
	for(i = 0; i < 371; i++)
	{
		check << segment[i].left.x << "  " << segment[i].left.y << "  " << segment[i].left.z << "  " << segment[i].right.x << "  " << segment[i].right.y << "  " << segment[i].right.z << 
			"  " << segment[i].radius << "  " << segment[i].type << "\n";
	}
	check.close();
	*/
	
	tdble maxK=0;

	
		
	
		
	for(i=0; i < 371; i++)
	{
		if(segment[i].type == TR_STR)
		{
			segment[i].K = 0;
		}
		else if(segment[i].type == TR_LFT)
		{
			segment[i].K = 1/segment[i].radius;
		}
		else if(segment[i].type == TR_RGT)
		{
			segment[i].K = -1/segment[i].radius;
		}
		if(abs(segment[i].K) > maxK)
		{
			maxK = abs(segment[i].K);
		}
	}
	
	srand((unsigned)time(0));

	std::ofstream outfile;
	outfile.open ("paths_found.txt");

	minEnergy = 9999;
	for(j=0; j<1; j++)
	{
		for(k=0; k<371; k++)
		{
			path_point_x[k] = (segment[k].left.x + segment[k].right.x)/2;
			path_point_y[k] = (segment[k].left.y + segment[k].right.y)/2;
		}
		//std::cout << path_point_x[0] << "  " << path_point_y[0] << "\n";
		//std::cout << path_point_x[1] << "  " << path_point_y[1] << "\n";
		Energy = 0;
		outfile << "path number  " << j << "\n";


		for(int i = 0; i <371; i++)
		{
			s = (segment[(i+2)%371].K - segment[(i+4)%371].K)/maxK;
			
			p = (float)rand()/(float)RAND_MAX;

			//std::cout << p <<"\n";

			if(s == 0)
			{
				x = 2*p-1;
			}
			else
			{
				x = (-1 + sqrt(1-s*(2-s-4*p)))/s;
			}
			x = (x + 1)/2;


			x1 = segment[(i+1)%371].left.x;
			y1 = segment[(i+1)%371].left.y;
			x2 = segment[(i+1)%371].right.x;
			y2 = segment[(i+1)%371].right.y;

			//std::cout << x1 << "  " << x2 << "  ";

			dx1 = x1 - x2;
			dy1 = y1 - y2;
			
			x3 = path_point_x[i];
			y3 = path_point_y[i];
			x4 = segment[i].left.x;
			y4 = segment[i].left.y;;

			dx2 = x1 - x4;
			dy2 = y1 - y4;
		
			if(dx1 == 0)
			{
				
				m2 = dy2/dx2;
				c2 = y3 - m2*x3;

				//std::cout << "11111 \n" << y3 << "  " << y4 << "  " << dy2<< "  " << dx2 << "  " << m2 << "  " << c2 << "\n";
				intersection_X = x1;
				intersection_Y = m2*x1 +c2;
			}
			else if(dx2 == 0)
			{
				//std::cout << "22222 \n";
				m1 = dy1/dx1;
				c1 = y1 - m1*x1;

				intersection_X = x3;
				intersection_Y = m1*x3 + c1;
			}
			else
			{
				//std::cout << "3333333 \n";
				m1 = dy1/dx1;
				c1 = y1 - m1*x1;
				m2 = dy2/dx2;
				c2 = y3 - m2*x3;
			
				intersection_X = (c2 - c1)/(m1 - m2);
				intersection_Y = m1*intersection_X +c1;
			}

			//std::cout << intersection_X << "  " << intersection_Y << "  " << m2 << "  " << c2 << "\n";

			//std::cout << intersection_X << "  " << x1 << "  " << x2 << "\n";

			if((x1 != x2) && (intersection_X > std::max(x1,x2)))
			{
				if(x1 > x2)
				{
					intersection_X = x1;
					intersection_Y = y1;
				}
				else
				{
					intersection_X = x2;
					intersection_Y = y2;
				}
			}
			else if((x1 != x2) && (intersection_X < std::min(x1,x2)))
			{
				
				if(x1 < x2)
				{
					intersection_X = x1;
					intersection_Y = y1;
				}
				else
				{
					intersection_X = x2;
					intersection_Y = y2;
				}
			}
			
			if((y1 != y2) && (intersection_Y > std::max(y1,y2)))
			{
				if(y1 > y2)
				{
					intersection_X = x1;
					intersection_Y = y1;
				}
				else
				{
					intersection_X = x2;
					intersection_Y = y2;
				}
			}
			else if((y1 != y2) && (intersection_Y < std::min(y1,y2)))
			{
				if(y1 < y2)
				{
					intersection_X = x1;
					intersection_Y = y1;
				}
				else
				{
					intersection_X = x2;
					intersection_Y = y2;
				}
			}

			//std::cout << intersection_X << "  " << intersection_Y << "\n";
			//std::cout << intersection_X << "  " << x1 << "  " << x2 << "\n";

			temp1 = intersection_X - x3;
			temp2 = intersection_Y - y3;
			temp1 = temp1*temp1;
			temp2 = temp2*temp2;
			temp3 = temp1 + temp2;
			temp3 = sqrt(temp3);
			distance = temp3*tan(15*3.14/180);

			//std::cout << intersection_X << "  " << x3 << "  " << distance << "\n";

			temp1 = x1 - intersection_X;
			temp2 = y1 - intersection_Y;
			temp1 = temp1*temp1;
			temp2 = temp2*temp2;
			temp3 = temp1 + temp2;
			temp3 = sqrt(temp3);
			distance_l = temp3;
			if(distance_l == 0)
	

			temp1 = x2 - intersection_X;
			temp2 = y2 - intersection_Y;
			temp1 = temp1*temp1;
			temp2 = temp2*temp2;
			temp3 = temp1 + temp2;
			temp3 = sqrt(temp3);
			distance_r = temp3;
		

			temp1 = x1 - x2;
			temp2 = y1 - y2;
			temp1 = temp1*temp1;
			temp2 = temp2*temp2;
			temp3 = temp1 + temp2;
			temp3 = sqrt(temp3);
			width = temp3;


			if((distance_l + distance_r) > width)
			{
				if(distance_l < distance_r)
				{
					intersection_X = x1;
					intersection_Y = y1;
				}
				else
				{
					intersection_X = x2;
					intersection_Y = y2;
				}
			}		

			//std::cout << intersection_X << "  " << intersection_Y << "\n";
			//std::cout << intersection_X << "  " << x3 << "  " << distance << "\n";

			temp1 = intersection_X - x3;
			temp2 = intersection_Y - y3;
			temp4 = temp1*temp1;
			temp5 = temp2*temp2;
			temp3 = temp4 + temp5;
			temp3 = sqrt(temp3);
			distance = temp3*5*3.14/180;

			temp1 = intersection_X - x1;
			temp2 = intersection_Y - y1;
			temp4 = temp1*temp1;
			temp5 = temp2*temp2;
			temp3 = temp4 + temp5;
			temp3 = sqrt(temp3);
			distance_l = temp3;
			if(distance_l == 0)
			{
				vector_xl = 0;
				vector_yl = 0;
			}
			else
			{
				vector_xl = temp1/distance_l;
				vector_yl = temp2/distance_l;
			}

			temp1 = intersection_X - x2;
			temp2 = intersection_Y - y2;
			temp4 = temp1*temp1;
			temp5 = temp2*temp2;
			temp3 = temp4 + temp5;
			temp3 = sqrt(temp3);
			distance_r = temp3;
			if(distance_r == 0)
			{
				vector_xr = 0;
				vector_yr = 0;
			}
			else
			{
			vector_xr = temp1/distance_r;
			vector_yr = temp2/distance_r;
			}

		
			//std::cout << distance << "  " << distance_l << "  " << distance_r << "\n";
/*
			
				if(abs(x1 - x2) < abs(y1 - y2) )
				{
					if(y1 < y2)
					{
						Y_left = intersection_Y - distance*vector_xl;
						Y_right = intersection_Y + distance*vector_xr;
						X_left = intersection_X - distance*vector_xl;
						X_right = intersection_X + distance*vector_xr;
					}
					else
					{
						Y_left = intersection_Y + distance*vector_xl;
						Y_right = intersection_Y - distance*vector_xr;
						X_left = intersection_X + distance*vector_xl;
						X_right = intersection_X - distance*vector_xr;
					}
				}
				else
				{
					if(x1 < x2)
					{
						X_left = intersection_X - distance*vector_xl;
						X_right = intersection_X + distance*vector_xr;
						Y_left = intersection_Y - distance*vector_xl;
						Y_right = intersection_Y + distance*vector_xr;
					}
					else
					{
						X_left = intersection_X + distance*vector_xl;
						X_right = intersection_X - distance*vector_xr;
						Y_left = intersection_Y + distance*vector_xl;
						Y_right = intersection_Y - distance*vector_xr;
					}
				}


			*/


			if(distance > distance_l)
			{
				X_left = x1;
				Y_left = y1;
			}
			else
			{
				X_left = intersection_X - (distance*vector_xl);
				Y_left = intersection_Y - (distance*vector_yl); 
			}

			if(distance > distance_r)
			{
				X_right = x2;
				Y_right = y2;
			}
			else
			{
				X_right = intersection_X - (distance*vector_xr);
				Y_right = intersection_Y - (distance*vector_yr);
			}
			
			//std::cout << intersection_Y << "  " << Y_left << "  " << y1 << "  " << distance << "  " << distance_l << "  " << vector_yl << "\n  ";

			

			//std::cout << x1 << "  " << X_left << "  " << X_right << "  " << x2 << "  " <<distance << "  " << distance_l << "  " << distance_r << "\n";

			//std::cout << x1 << "  " << X_right << "  " << x2 << "  " << y1 << "  " << Y_right << "  " << y2 << "  " <<distance << "  " << distance_r << "\n";
	
			
			//std :: cout << i << "  " << X_left << "  " << Y_left << "  " << X_right << "  " << Y_right << "\n";

			path_point_x[(i+1)%371] = std::min(X_left,X_right) + x*abs(X_left - X_right);
			path_point_y[(i+1)%371] = std::min(Y_left,Y_right) + x*abs(Y_left - Y_right);

			
			//std::cout << intersection_X << "  " << intersection_Y << "\n";
			//std :: cout << i << "  " << x1 << "  " << path_point_x[(i+1)%371] << "  " << x2 << "  " << y1 << "  " << path_point_y[(i+1)%371] << "  " << y2 << "\n";
			//std :: cout << i << "  ";
			Energy = Energy + race_line_curvature(path_point_x[abs(i-1)], path_point_y[abs(i-1)], path_point_x[i], path_point_y[i], path_point_x[(i+1)%371], path_point_y[(i+1)%371]);

			outfile << X_left << "  " << Y_left << "  " << X_right << "  "  << Y_right << "  " << intersection_X << "  " << intersection_Y << "  " << path_point_x[(i+1)%371] << "  " << path_point_y[(i+1)%371] << "  " << segment[(i+1)%371].left.x << "  " << segment[(i+1)%371].left.y 
				<< "  " << segment[(i+1)%371].right.x << "  " << segment[(i+1)%371].right.y << "  " << "\n";

		}

		outfile << "energy of this line  " << Energy << "\n";
		if(Energy < minEnergy)
		{
			minEnergy = Energy;
		}
			
	}

	outfile << "minimum energy computed =  " << minEnergy << "\n";
	outfile.close();

	std::cout << "end of program";
	getchar();

}

float race_line_curvature(float prev_x, float prev_y, float curr_x, float curr_y, float next_x, float next_y)
{
	float curvature = 0;

	float yDelta_a = curr_y - prev_y;
    float xDelta_a = curr_x - prev_x;
    float yDelta_b = next_y - curr_y;
    float xDelta_b = next_x - curr_x;
 
	if(xDelta_a !=0 && xDelta_b != 0)
	{
		float aSlope = yDelta_a/xDelta_a;
		float bSlope = yDelta_b/xDelta_b;  
    
		if(aSlope == 0 || bSlope == 0 || (bSlope-aSlope) == 0)
		{
			curvature = 0;
		}
		else if(aSlope != bSlope)
		{
			float center_x = (aSlope*bSlope*(prev_y - next_y) + bSlope*(prev_x + curr_x) - aSlope*(curr_x + next_x) )/(2* (bSlope-aSlope) );
			float center_y = -1*(center_x - (prev_x + curr_x)/2)/aSlope +  (prev_y + curr_y)/2;

			float temp1 = center_x - curr_x;
			float temp2 = center_y - curr_y;
			temp1 = temp1*temp1;
			temp2 = temp2*temp2;
			float temp3 = temp1 + temp2;
			temp3 = sqrt(temp3);
	
			float radius = temp3;
			curvature = 1/radius;
		}
	}

	//std::cout << curvature << "\n";
	return (curvature);
}